/*
 * ADTPro - Apple Disk Transfer ProDOS
 * Copyright (C) 2006 by David Schmidt
 * david__schmidt at users.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the 
 * Free Software Foundation; either version 2 of the License, or (at your 
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.adtpro;

import java.io.*;
import java.net.InetAddress;
import java.net.SocketException;
import java.util.GregorianCalendar;

import org.adtpro.resources.Messages;
import org.adtpro.transport.ATransport;
import org.adtpro.transport.SerialTransport;
import org.adtpro.transport.TransportTimeoutException;
import org.adtpro.transport.UDPTransport;

import org.adtpro.disk.Disk;
import org.adtpro.gui.Gui;
import org.adtpro.utilities.UnsignedByte;

public class CommsThread extends Thread
{
  private boolean _shouldRun = true;

  private ATransport _transport;

  private Gui _parent;

  private static final byte ACK = 0x06, NAK = 0x15;

  private int[] CRCTABLE = new int[256];

  private int _maxRetries = 15;

  GregorianCalendar _startTime = null, _endTime = null;

  private float _diffMillis;

  private Worker _worker = null;

  public CommsThread(Gui parent, String one, String two)
  {
    // System.out.println("CommsThread constructor.");
    _parent = parent;
    try
    {
      if (one.equals(Messages.getString("Gui.Ethernet")))
      {
        _transport = (ATransport) new UDPTransport("6502");
        _transport.open();
        _parent.setTitle(Messages.getString("Gui.EthTitle") + " " + InetAddress.getLocalHost().getHostAddress());
      }
      else
        _transport = (ATransport) new SerialTransport(one, two);
    }
    catch (Exception ex)
    {
      System.out.println("CommsThread constructor exception: " + ex);
      _shouldRun = false;
    }
  }

  public void run()
  {
    if (_shouldRun)
    {
      makeCrcTable();
      commandLoop();
    }
    else
      _parent.cancelCommsThread();
  }

  public void commandLoop()
  {
    // System.out.println("CommsThread.commandLoop() starting.");
    byte oneByte = (byte) 0x00;
    while (_shouldRun)
    {
      // System.out.println("DEBUG: CommsThread.commandLoop() Waiting for
      // command from Apple."); //$NON-NLS-1$
      oneByte = waitForData();
      if (_shouldRun)
      {
        // System.out.println("DEBUG: CommsThread.commandLoop() Received a byte:
        // " + oneByte); //$NON-NLS-1$
        _parent.setProgressMaximum(0);
        switch (oneByte)
        {
          case (byte) 195: // "C": CD
            _parent.setMainText(Messages.getString("CommsThread.2")); //$NON-NLS-1$
            _parent.setSecondaryText(""); //$NON-NLS-1$
            // System.out.println("CD..."); //$NON-NLS-1$
            changeDirectory();
            _parent.setSecondaryText(_parent.getWorkingDirectory()); //$NON-NLS-1$
            break;
          case (byte) 196: // "D": DIR
            _parent.setMainText(Messages.getString("CommsThread.1")); //$NON-NLS-1$
            _parent.setSecondaryText(_parent.getWorkingDirectory()); //$NON-NLS-1$
            // System.out.println("Dir..."); //$NON-NLS-1$
            sendDirectory();
            _parent.setSecondaryText(_parent.getWorkingDirectory()); //$NON-NLS-1$
            break;
          case (byte) 208: // "P": Put (Send)
            _parent.setMainText(Messages.getString("CommsThread.16")); //$NON-NLS-1$
            _parent.setSecondaryText(""); //$NON-NLS-1$
            // System.out.println("Put/Send..."); //$NON-NLS-1$
            receiveDisk(false);
            break;
          case (byte) 199: // "G": Get (Receive)
            _parent.setMainText(Messages.getString("CommsThread.3")); //$NON-NLS-1$
            _parent.setSecondaryText(""); //$NON-NLS-1$
            // System.out.println("Get/Receive..."); //$NON-NLS-1$
            sendDisk();
            break;
          case (byte) 194: // "B": Batch send
            _parent.setMainText(Messages.getString("CommsThread.3")); //$NON-NLS-1$
            _parent.setSecondaryText(""); //$NON-NLS-1$
            // System.out.println("Get/Receive..."); //$NON-NLS-1$
            receiveDisk(true);
            break;
          case (byte) 217: // "P": Ping
            _parent.setSecondaryText("Ping received from client."); //$NON-NLS-1$
            // System.out.println("Ping..."); //$NON-NLS-1$
            _transport.pushBuffer();
            _transport.flushReceiveBuffer();
            break;
          case (byte) 218: // "Q": Size
            _parent.setMainText(Messages.getString("CommsThread.14")); //$NON-NLS-1$
            _parent.setSecondaryText(""); //$NON-NLS-1$
            // System.out.println("queryFileSize..."); //$NON-NLS-1$
            queryFileSize();
            break;
          case (byte) 210: // "R": Receive (Legacy ADT style)
            _parent.setMainText(Messages.getString("CommsThread.11")); //$NON-NLS-1$
            _parent.setSecondaryText(""); //$NON-NLS-1$
            // System.out.println("Legacy receive..."); //$NON-NLS-1$
            send140kDisk();
            break;
          case (byte) 211: // "S": Send (Legacy ADT style)
            _parent.setMainText(Messages.getString("CommsThread.15")); //$NON-NLS-1$
            _parent.setSecondaryText(""); //$NON-NLS-1$
            // System.out.println("Legacy send..."); //$NON-NLS-1$
            receive140kDisk();
            break;
          default:
            // System.out.println("not understood... received: " +
            // UnsignedByte.toString(oneByte));
            break;
        }
      }
    }
    // System.out.println("CommsThread.commandLoop() ending.");
  }

  public void sendDirectory()
  {
    int i, j, line;
    try
    {
      _transport.writeBytes("DIRECTORY OF "); //$NON-NLS-1$
      _transport.writeBytes(_parent.getWorkingDirectory());
      if (((_parent.getWorkingDirectory().length() + 13) % 40) != 0) _transport.writeByte('\r');
      // 40 dashes separates the wheat from the chaff
      _transport.writeBytes("----------------------------------------"); //$NON-NLS-1$

      File[] files = _parent.getFiles();
      if (files.length > 0)
      {
        i = 0;
        j = 0;
        line = (_parent.getWorkingDirectory().length() + 13) / 40 + 2;
        while (_shouldRun)
        {
          if ((i > 0) && (i + files[j].getName().length() > 40))
          {
            while (i++ < 40)
              _transport.writeByte(' ');
            i = 0;
            line++;
          }
          if (line > 20)
          {
            line = 0;
            _transport.writeByte('\0');
            _transport.writeByte('\1');
            _transport.pushBuffer();
            if (waitForData() == '\0') break;
          }
          line += (files[j].getName().length() / 40);
          i += (files[j].getName().length() % 40);
          _transport.writeBytes(files[j].getName());
          j++;
          if (j + 1 > files.length)
          {
            break;
          }
          do
          {
            if (i == 40)
            {
              i = 0;
              line++;
              break;
            }
            _transport.writeByte(' ');
          }
          while ((++i % 14) > 0);
        }
      }
      else
        _transport.writeBytes("NO FILES"); //$NON-NLS-1$
      _transport.writeByte('\0');
      _transport.writeByte('\0');
      _transport.pushBuffer();
    }
    catch (Throwable t1)
    {
      System.out.println("sendDirectory exception:"); //$NON-NLS-1$
      System.out.println(t1);
      _transport.writeBytes("NO FILES"); //$NON-NLS-1$
      _transport.writeByte('\0');
      _transport.writeByte('\0');
      _transport.pushBuffer();
    }
  }

  public void queryFileSize()
  {
    long length = 0;
    byte sizeLo = 0, sizeHi = 0, rc = (byte) 0xff;
    String requestedFileName = receiveName();
    Disk disk = null;

    try
    {
      disk = new Disk(_parent.getWorkingDirectory() + File.separator + requestedFileName);
      // System.out.println("queryFileSize found file " +
      // _parent.getWorkingDirectory() + File.separator + requestedFileName);
      // //$NON-NLS-1$
    }
    catch (IOException e)
    {
      try
      {
        disk = new Disk(requestedFileName);
        // System.out.println("queryFileSize found file " + requestedFileName);
        // //$NON-NLS-1$
      }
      catch (IOException e2)
      {
        // System.out.println("can't read file: " + requestedFileName + ".");
        // //$NON-NLS-1$ //$NON-NLS-2$
        rc = 0x02; // Unable to open file
      }
    }
    catch (ArrayIndexOutOfBoundsException ix)
    {
      rc = 0x04; // Unrecognized file format
    }
    if (disk == null)
    {
      if (rc == (byte) 0xff) rc = 0x02; // Unable to open file
      // else let rc be whatever it was before
    }
    else
      if (disk.getImageOrder() == null) rc = 0x04; // Unrecognized file format
      else
      {
        length = disk.getImageOrder().getBlocksOnDevice();
        sizeLo = UnsignedByte.loByte(length);
        // System.out.println("loByte of " + requestedFileName + " is: " +
        // UnsignedByte.intValue(sizeLo)); //$NON-NLS-1$ //$NON-NLS-2$
        sizeHi = UnsignedByte.hiByte(length);
        // System.out.println("hiByte of " + requestedFileName + " is: " +
        // UnsignedByte.intValue(sizeHi)); //$NON-NLS-1$ //$NON-NLS-2$
        rc = 0;
      }

    if (disk != null) _parent.setSecondaryText(disk.getFilename());
    else
      _parent.setSecondaryText(requestedFileName);
    // System.out.println("queryFileSize lo:" + UnsignedByte.toString(sizeLo) +
    // " hi:" + UnsignedByte.toString(sizeHi)); //$NON-NLS-1$ //$NON-NLS-2$
    _transport.writeByte(sizeLo);
    _transport.writeByte(sizeHi);
    _transport.writeByte(rc);
    _transport.pushBuffer();
  }

  public void changeDirectory()
  {
    byte rc = 0x06;

    String requestedDirectory = receiveName();
    if (_shouldRun)
    {
      rc = _parent.setWorkingDirectory(requestedDirectory);
      _transport.writeByte(rc);
      _transport.pushBuffer();
    }
  }

  public void receiveDisk(boolean generateName)
  /* Main receive routine - Host <- Apple (Apple sends) */
  {
    // System.out.println("DEBUG: receiveDisk() entry.");
    _startTime = new GregorianCalendar();
    // System.out.print("Waiting for name..."); //$NON-NLS-1$
    String name = _parent.getWorkingDirectory() + File.separator + receiveName();
    // System.out.println(" received name: " + name); //$NON-NLS-1$
    File f = null;
    String nameGen, zeroPad;
    if (generateName)
    {
      do
      {
        if (lastFileNumber < 10) zeroPad = "000";
        else
          if (lastFileNumber < 100) zeroPad = "00";
          else
            if (lastFileNumber < 1000) zeroPad = "0";
            else
              zeroPad = "";
        nameGen = zeroPad + lastFileNumber;
        f = new File(name + nameGen + ".PO");
        lastFileNumber++;
      }
      while (f.exists());
      name = name + nameGen + ".PO";
    }
    else
      f = new File(name);
    FileOutputStream fos = null;
    byte[] buffer = new byte[20480];
    int part, length, packetResult = 0;
    byte report, sizelo, sizehi;
    int halfBlock;
    int blocksDone = 0;

    // New ADT protcol - file size to expect
    // System.out.print("Waiting for sizeLo..."); //$NON-NLS-1$
    sizelo = waitForData();
    // System.out.println(" received sizeLo: " + UnsignedByte.intValue(sizelo));
    // //$NON-NLS-1$
    // System.out.print("Waiting for sizeHi..."); //$NON-NLS-1$
    sizehi = waitForData();
    // System.out.println(" received sizeHi: " + UnsignedByte.intValue(sizehi));
    // //$NON-NLS-1$
    length = UnsignedByte.intValue(sizelo, sizehi);
    _parent.setProgressMaximum(length);
    try
    {
      fos = new FileOutputStream(f);
      // ready for transfer
      _transport.writeByte(0x00);
      _transport.pushBuffer();
      // System.out.println("DEBUG: receiveDisk() about to wait for ACK from
      // apple...");
      if (waitForData() == ACK)
      {
        // System.out.println("DEBUG: receiveDisk() received ACK from apple.");
        _parent.setProgressMaximum((int) length * 2); // Half-blocks
        _parent.setSecondaryText(name);
        int numParts = (int) length / 40;
        int remainder = (int) length % 40;
        for (part = 0; part < numParts; part++)
        {
          // System.out.println("Receiving part " + (part + 1) + " of " +
          // numParts + "; "); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
          for (halfBlock = 0; halfBlock < 80; halfBlock++)
          {
            packetResult = receivePacket(buffer, halfBlock * 256, (part * 80 + halfBlock));
            if (packetResult != 0) break;
            blocksDone++;
            _parent.setProgressValue(blocksDone);
          }
          if (packetResult != 0) break;
          // System.out.println("Writing part " + (part + 1) + " of " +
          // numParts + "."); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
          fos.write(buffer);
        }
        // System.out.println("Bottom of for loop... packetResult:
        // "+packetResult);
        if ((packetResult == 0) && (remainder > 0))
        {
          // System.out.println(" ... read " + charsRead + " chars.");
          // System.out.println("Receiving remainder part."); //$NON-NLS-1$
          for (halfBlock = 0; halfBlock < (remainder * 2); halfBlock++)
          {
            packetResult = receivePacket(buffer, halfBlock * 256, (part * 80 + halfBlock));
            if (packetResult == -1) break;
            blocksDone++;
            _parent.setProgressValue(blocksDone);
          }
          if (packetResult == 0)
          {
            // System.out.println("Writing remainder " + remainder + "
            // blocks."); //$NON-NLS-1$ //$NON-NLS-2$
            fos.write(buffer, 0, remainder * 512);
          }
        }
        fos.close();
      }
      else
      {
        packetResult = -1;
      }
      if (packetResult == 0)
      {
        report = waitForData();
        _endTime = new GregorianCalendar();
        _diffMillis = (float) (_endTime.getTimeInMillis() - _startTime.getTimeInMillis()) / (float) 1000;
        if (report == 0x00)
        {
          _parent.setSecondaryText(Messages.getString("CommsThread.19") + " in " + _diffMillis + " seconds.");
          System.out.println("Apple sent disk image " + name + " successfully in "
              + (float) (_endTime.getTimeInMillis() - _startTime.getTimeInMillis()) / (float) 1000 + " seconds.");
        }
        else
        {
          _parent.setSecondaryText(Messages.getString("CommsThread.20") + " in " + _diffMillis + " seconds.");
          System.out.println("Apple sent disk image " + name + " with errors."); //$NON-NLS-1$ //$NON-NLS-2$
        }
      }
      else
      {
        _parent.setSecondaryText(Messages.getString("CommsThread.21"));
        _parent.clearProgress();
        _transport.flushReceiveBuffer();
        _transport.flushSendBuffer();
      }
    }
    catch (FileNotFoundException ex)
    {
      _transport.writeByte(0x02); // New ADT protocol: HMFIL - unable to write
      // file
      _transport.pushBuffer();
    }
    catch (IOException ex2)
    {
      _transport.writeByte(0x02); // New ADT protocol: HMFIL - unable to write
      // file
      _transport.pushBuffer();
    }
    finally
    {
      if (fos != null) try
      {
        fos.close();
      }
      catch (IOException io)
      {}
    }
    // System.out.println("DEBUG: receiveDisk() exit.");
  }

  public void sendDisk()
  /* Main send routine - Host -> Apple (Host sends) */
  {
    // System.out.println("DEBUG: sendDisk() entry.");
    byte[] buffer = new byte[Disk.BLOCK_SIZE];
    int halfBlock, blocksDone = 0;
    byte ack, report;
    int length;
    boolean sendSuccess = false;
    _startTime = new GregorianCalendar();
    /*
     * ADT protocol: receive the requested file name
     */
    String name = receiveName();

    Disk disk = null;
    try
    {
      disk = new Disk(_parent.getWorkingDirectory() + File.separator + name);
    }
    catch (IOException io)
    {
      try
      {
        disk = new Disk(name);
      }
      catch (IOException io2)
      {}
    }
    if (disk != null)
    {
      if (disk.getImageOrder() != null)
      {
        // If the file exists, then...
        _transport.writeByte(0x00); // Tell Apple ][ we're ready to go
        _transport.pushBuffer();
        // System.out.println("DEBUG: sendDisk() about to wait for initial
        // ack.");
        ack = waitForData();
        // System.out.println("Received initial reply from Apple: " + ack);
        // //$NON-NLS-1$
        if (ack == 0x06)
        {
          length = disk.getImageOrder().getBlocksOnDevice();
          _parent.setProgressMaximum(length * 2); // Half-blocks
          _parent.setSecondaryText(disk.getFilename());
          // System.out.println("Length is " + length + " blocks.");
          // //$NON-NLS-1$ //$NON-NLS-2$
          for (int block = 0; block < length; block++)
          {
            buffer = disk.readBlock(block);
            for (halfBlock = 0; halfBlock < 2; halfBlock++)
            {
              sendSuccess = sendPacket(buffer, halfBlock * 256);
              if (sendSuccess)
              {
                blocksDone++;
                _parent.setProgressValue(blocksDone);
              }
              else
                break;
            }
            if (!sendSuccess) break;
          }
          if (sendSuccess)
          {
            report = waitForData();
            _endTime = new GregorianCalendar();
            _diffMillis = (float) (_endTime.getTimeInMillis() - _startTime.getTimeInMillis()) / (float) 1000;
            if (report == 0x00)
            {
              _parent.setSecondaryText(Messages.getString("CommsThread.17") + " in " + _diffMillis + " seconds.");
              System.out
                  .println("Apple received disk image " + name + " successfully in " + (float) (_endTime.getTimeInMillis() - _startTime.getTimeInMillis()) / (float) 1000 + " seconds."); //$NON-NLS-1$ //$NON-NLS-2$
            }
            else
            {
              _parent.setSecondaryText(Messages.getString("CommsThread.18"));
              System.out.println("Apple received disk image " + name + " with " + report + " errors."); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
            }
          }
          else
          {
            _parent.setSecondaryText(Messages.getString("CommsThread.21"));
            _parent.clearProgress();
            _transport.flushReceiveBuffer();
            _transport.flushSendBuffer();
          }
        }
        else
        {
          // System.out.print("No ACK received from the Apple...");
          // //$NON-NLS-1$
          _parent.setSecondaryText(Messages.getString("CommsThread.21"));
          _parent.clearProgress();
          _transport.flushReceiveBuffer();
          _transport.flushSendBuffer();
        }
      }
      else
      {
        // New ADT protocol: HMFIL - can't open the file
        _transport.writeByte(0x02);
        _transport.pushBuffer();
      }
    }
    else
    {
      // New ADT protocol: HMFIL - can't open the file
      _transport.writeByte(0x02);
      _transport.pushBuffer();
    }
    // System.out.println("DEBUG: sendDisk() exit.");
  }

  /**
   * send140kDisk - legacy ADT protocol send 140k disk function
   * 
   * Note these images will be coming in and going out in DOS order.
   */
  public void send140kDisk()
  {
    _startTime = new GregorianCalendar();
    /*
     * ADT protocol: receive the requested file name
     */
    String name = receiveName();
    byte ack;
    int bufSize = 28672;
    byte[] buffer = new byte[bufSize];
    File f = new File(name);
    int rc = 0;
    int sectorsDone = 0;
    FileInputStream fis = null;
    boolean sendSuccess = false;

    if (!f.isFile())
    {
      f = new File(_parent.getWorkingDirectory() + File.separator + name);
      if (!f.isFile())
      {
        _transport.writeByte(26); // ADT protocol - can't open
        _transport.pushBuffer();
        rc = -1;
      }
    }
    if (rc == 0)
    {
      long length = f.length();
      if (length != (long) 143360)
      {
        /*
         * ADT protocol: send error (message) number
         */
        // System.out.println("Not a 140k image"); //$NON-NLS-1$
        _transport.writeByte(30); // ADT protocol - not a 140k image
        _transport.pushBuffer();
        rc = -1;
      }
      else
      {
        /*
         * ADT protocol: send trigger
         */
        _parent.setSecondaryText(name);
        // If the file exists, is pristine, etc., then...
        _transport.writeByte(0x00);
        _transport.pushBuffer();
        try
        {
          /*
           * ADT protocol: receive acknowledgement for "previous" sector
           */
          ack = waitForData();
          if (ack == 0x06)
          {
            _parent.setProgressMaximum(560); // Sectors
            fis = new FileInputStream(f);
            for (int part = 0; part < 5; part++)
            {
              // System.out.print("Reading track " + track);
              // System.out.println(" ... read " + charsRead + " chars.");
              int charsRead = fis.read(buffer);
              for (int track = 0; track < 7; track++)
              {
                for (int sector = 15; sector >= 0; sector--)
                {
                  // System.out.println("Sending track " + (track + (part * 7))
                  // + " sector " + sector + "."); //$NON-NLS-1$ //$NON-NLS-2$
                  // //$NON-NLS-3$
                  sendSuccess = sendPacket(buffer, (track * 4096 + sector * 256));
                  if (!sendSuccess) break;
                  sectorsDone++;
                  _parent.setProgressValue(sectorsDone);
                }
                if (!sendSuccess) break;
              }
              if (!sendSuccess) break;
            }
            fis.close();
            if (sendSuccess)
            {
              /*
               * ADT protocol: receive final error report
               */
              byte report = waitForData();
              _endTime = new GregorianCalendar();
              _diffMillis = (float) (_endTime.getTimeInMillis() - _startTime.getTimeInMillis()) / (float) 1000;
              if (report == 0x00)
              {
                _parent.setSecondaryText(Messages.getString("CommsThread.19") + " in " + _diffMillis + " seconds.");
                System.out
                    .println("Apple sent disk image " + name + " successfully in " + (float) (_endTime.getTimeInMillis() - _startTime.getTimeInMillis()) / (float) 1000 + " seconds."); //$NON-NLS-1$ //$NON-NLS-2$
              }
              else
              {
                _parent.setSecondaryText(Messages.getString("CommsThread.20") + " in " + _diffMillis + " seconds.");
                System.out.println("Apple sent disk image " + name + " with errors."); //$NON-NLS-1$ //$NON-NLS-2$
              }
            }
            else
            {
              _parent.setSecondaryText(Messages.getString("CommsThread.21"));
              _transport.flushReceiveBuffer();
              _transport.flushSendBuffer();
            }
          }
          // else
          // System.out.print("No ACK received from the Apple."); //$NON-NLS-1$
        }
        catch (IOException ex)
        {
          // System.out.print("Disk send aborted."); //$NON-NLS-1$
          _parent.setSecondaryText(Messages.getString("CommsThread.21"));
          _transport.flushReceiveBuffer();
          _transport.flushSendBuffer();
        }
        
        {
          if (fis != null) try
          {
            fis.close();
          }
          catch (IOException io)
          {

          }
        }
      }
    }
    // System.out.println("DEBUG: send140kDisk() exit.");
  }

  public boolean sendPacket(byte[] buffer, int offset)
  {
    boolean rc = false;

    int byteCount, crc, ok = NAK, currentRetries = 0;
    byte data, prev, newprev;

    // System.out.println("CommsThread.sendPacket() entry; offset "+offset+".");
    do
    {
      prev = 0;
      for (byteCount = 0; byteCount < 256;)
      {
        newprev = buffer[offset + byteCount];
        data = (byte) (UnsignedByte.intValue(newprev) - UnsignedByte.intValue(prev));
        prev = newprev;
        _transport.writeByte(data);
        if (UnsignedByte.intValue(data) > 0) byteCount++;
        else
        {
          while ((_shouldRun == true) && byteCount < 256 && buffer[offset + byteCount] == newprev)
          {
            byteCount++;
          }
          _transport.writeByte((byte) (byteCount & 0xFF)); // 256 becomes 0
        }
        if (!_shouldRun)
        {
          rc = false;
          break;
        }
      }
      if (_shouldRun)
      {
        crc = doCrc(buffer, offset, 256);
        _transport.writeByte((byte) (crc & 0xff));
        _transport.writeByte((byte) (((crc & 0xff00) >> 8) & 0xff));
        _transport.pushBuffer();
        // System.out.println("");
        // System.out.println("Locally calculated CRC: " + (crc & 0xffff));
        ok = waitForData();
        //We can't time out here!  This timeout has to be initiated from the Apple...
        //catch (TransportTimeoutException te)
        //{
        //  System.out.println("CommsThread.sendPacket() timeout.");
        //  ok = NAK;
        //}
        // System.out.println("ack from Apple: " + ok);
        if (ok == ACK) rc = true;
        else
          currentRetries++;
      }
    }
    while ((ok != ACK) && (_shouldRun == true) && (currentRetries < _maxRetries));

    // System.out.println("CommsThread.sendPacket() exit, rc = "+rc);
    return rc;
  }

  public void receive140kDisk()
  {
    _startTime = new GregorianCalendar();
    String name = _parent.getWorkingDirectory() + File.separator + receiveName();
    File f = new File(name);
    FileOutputStream fos = null;
    byte[] buffer = new byte[28672];
    int i, part, track, sector, packetResult = -1;
    int sectorsDone = 0;
    byte report;

    if (f.exists())
    {
      _transport.writeByte(0x1c); // ADT protocol - file exists
      _transport.pushBuffer();
    }
    else
    {
      try
      {
        fos = new FileOutputStream(f);
      }
      catch (FileNotFoundException ex)
      {
        // We expect a file not found exception
      }
      if (fos != null)
      {
        _parent.setSecondaryText(name);
        for (i = 0; i < buffer.length; i++)
          buffer[i] = 0x00;
        try
        {
          for (i = 0; i < 7; i++)
            fos.write(buffer);
          fos.close();
          _transport.writeByte(0x00); // File is now ready
          _transport.pushBuffer();
          fos = new FileOutputStream(f);
          while (waitForData() != ACK)
          {
            // TODO: What needs to happen here? Original ADT talked about
            // a bad header message...
            System.out.println("hrm, not getting an ACK from the Apple..."); //$NON-NLS-1$
          }
          _parent.setProgressMaximum(560); // sectors
          for (part = 0; part < 5; part++)
          {
            for (track = 0; track < 7; track++)
            {
              for (sector = 15; sector >= 0; sector--)
              {
                packetResult = receivePacket(buffer, (track * 4096) + (sector * 256), -1);
                // System.out.println("Received track " + ((7 * part) + track) + ", sector " + (15 - sector) + ": " + (receiveSuccess ? "Success." : "Failure."));
                if (packetResult != 0) break;
                sectorsDone++;
                _parent.setProgressValue(sectorsDone);
              }
              if (packetResult != 0) break;
            }
            fos.write(buffer);
            if (packetResult != 0) break;
          }
          fos.close();
          if (packetResult == 0)
          {
            report = waitForData();
            _endTime = new GregorianCalendar();
            _diffMillis = (float) (_endTime.getTimeInMillis() - _startTime.getTimeInMillis()) / (float) 1000;
            if (report == 0x00)
            {
              _parent.setSecondaryText(Messages.getString("CommsThread.19") + " in " + _diffMillis + " seconds.");
              System.out
                  .println("Apple sent disk image " + name + " successfully in " + (float) (_endTime.getTimeInMillis() - _startTime.getTimeInMillis()) / (float) 1000 + " seconds."); //$NON-NLS-1$ //$NON-NLS-2$
            }
            else
            {
              _parent.setSecondaryText(Messages.getString("CommsThread.20") + " in " + _diffMillis + " seconds.");
              // System.out.println("Received disk image " + name + " with
              // errors."); //$NON-NLS-1$ //$NON-NLS-2$
            }
          }
          else
          {
            _parent.setSecondaryText(Messages.getString("CommsThread.21"));
            _transport.flushReceiveBuffer();
            _transport.flushSendBuffer();
          }
        }
        catch (IOException ex2)
        {
          _transport.writeByte(0x1a); // ADT protocol - unable to write file
          _transport.pushBuffer();
        }
        finally
        {
          if (fos != null) try
          {
            fos.close();
          }
          catch (IOException io)
          {}
        }
      }
    }
    // System.out.println("DEBUG: receive140kDisk() exit.");
  }

  public int receivePacket(byte[] buffer, int offset, int buffNum)
  // Receive a packet with RLE compression
  // Returns:
  // 0 on successful read - block/halfblock numbers, CRC matched
  // -1 on inability to read a packet successfully (timeouts, retries exhausted)
  {
    int byteCount, retries = 0;
    int received_crc = -1, computed_crc = 0;
    int incomingBlockNum, incomingHalf, blockNum, halfNum;
    byte data = 0x00, prev, crc1 = 0, crc2 = 0;
    int rc = 0;
    boolean restarting = false;

    // System.out.println("DEBUG: receivePacket() entry; offset "+offset+", buffNum = "+buffNum+".");
    do
    {
      // System.out.println(" top of receivePacket loop.");
      rc = 0;
      prev = 0;
      restarting = false;
      if (_transport.hasPreamble())
      {
        try
        {
          // Wait for the block number...
          incomingBlockNum = UnsignedByte.intValue(waitForData2(5000));
          incomingBlockNum = incomingBlockNum + ((UnsignedByte.intValue(waitForData2(1000)) * 256));
          data = waitForData2(1000);
          incomingHalf = Math.abs(2 - data); // Get the half block

          blockNum = buffNum / 2;
          halfNum = buffNum % 2;
          //System.out.println("BlockNum: " + blockNum + " Incoming: " + incomingBlockNum + " HalfNum: " + halfNum + " Incoming: " + incomingHalf);
          if ((incomingBlockNum != blockNum) || (incomingHalf != halfNum))
          {
            if ((incomingBlockNum == (blockNum - 1) && (incomingHalf == (1 - halfNum))))
            {
              rc = -2;
              System.out.println("Block numbers were close (full); acknowledging.");
            }
            else if ((incomingBlockNum == blockNum) && (incomingHalf == (1 - halfNum)))
            {
              rc = -2;
              System.out.println("Block numbers were close (half); acknowledging.");
            }
            else
            {
              rc = -1;
              System.out.println("Block numbers didn't match.");
            }
          }
          else
            rc = 0;
        }
        catch (TransportTimeoutException tte)
        {
          rc = -1;
          System.out.println("TransportTimeoutException! (location 1)");
        }
      }
      if (rc == 0)
      {
        for (byteCount = 0; byteCount < 256;)
        {
          try
          {
            // Wait for a byte...
            data = waitForData2(1000);
            if (UnsignedByte.intValue(data) > 0)
            {
              prev += UnsignedByte.intValue(data);
              // if (byteCount % 32 == 0) System.out.println("");
              buffer[offset + byteCount++] = prev;
              // System.out.print(UnsignedByte.toString(buffer[(offset + byteCount)-1]) + " ");
            }
            else
            {
              data = waitForData2(1000);
              do
              {
                // if (byteCount % 32 == 0) System.out.println("");
                buffer[offset + byteCount++] = prev;
                // System.out.print(UnsignedByte.toString(buffer[offset + byteCount - 1]) + " ");
              }
              while (_shouldRun && byteCount < 256 && byteCount != UnsignedByte.intValue(data));
            }
            if (!_shouldRun)
            {
              rc = -1;
              break;
            }
          }
          catch (TransportTimeoutException tte)
          {
            rc = -1;
            System.out.println("TransportTimeoutException! (location 2)");
            break;
          }
        }
        if (_shouldRun && !restarting)
        {
          // System.out.println("");
          // System.out.print("Receiving CRC bytes...");
          crc1 = waitForData();
          crc2 = waitForData();
          received_crc = UnsignedByte.intValue(crc1, crc2);
          computed_crc = doCrc(buffer, offset, 256);
          if (received_crc != computed_crc)
          {
            rc = -1;
            System.out.println("Incorrect CRC. Computed: " + computed_crc + " Received: " + received_crc); //$NON-NLS-1$ //$NON-NLS-2$
          }
          else
          {
            // System.out.println("Correct CRC. Computed: " + computed_crc + " Received: " + received_crc); //$NON-NLS-1$ //$NON-NLS-2$
            rc = 0;
          }
        }
      }
      if (rc == 0)
      {
        _transport.writeByte(ACK);
        _transport.pushBuffer();
      }
      else if (rc == -2)
      {
        /*
         *  We received an out-of-sync packet (likely a duplicate).
         *  Acknowledge it, and swing around again for another try.
         */
        _transport.writeByte(ACK);
        _transport.pushBuffer();
        _transport.flushReceiveBuffer();
        _transport.flushSendBuffer();
        retries++;
      }
      else
      {
        _transport.flushReceiveBuffer();
        _transport.flushSendBuffer();
        _transport.writeByte(NAK);
        _transport.pushBuffer();
        retries++;
      }
    }
    while ((rc != 0) && (_shouldRun == true) && (retries < _maxRetries));
    // System.out.println("DEBUG: receivePacket() exit.");

    return rc;
  }

  public byte waitForData()
  {
    byte oneByte = 0;
    try
    {
      oneByte = waitForData2(0);
    }
    catch (TransportTimeoutException e)
    {

    }
    return oneByte;
  }

  public byte waitForData2(int timeout) throws TransportTimeoutException
  {
    byte oneByte = 0;
    boolean readYet = false;
    while ((readYet == false) && (_shouldRun == true))
    {
      try
      {
        if (_transport != null)
        {
          oneByte = _transport.readByte(timeout);
          readYet = true;
        }
        else
        {
          _shouldRun = false;
        }
      }
      catch (TransportTimeoutException tte)
      {
        System.out.println("TransportTimeoutException!");
        throw tte;
        // _shouldRun = false;
      }
      catch (SocketException se)
      {
        _shouldRun = false;
      }
      catch (IOException ex)
      {
        ex.printStackTrace();
      }
      catch (Exception e)
      {
        e.printStackTrace();
      }
    }
    return oneByte;
  }

  public String receiveName()
  {
    byte oneByte;
    StringBuffer buf = new StringBuffer();

    for (int i = 0; i < 256; i++)
    {
      if (!_shouldRun) break;
      oneByte = waitForData();
      if (oneByte != (byte) 0x00)
      {
        buf.append((char) (UnsignedByte.intValue(oneByte) & 0x7f));
      }
      else
        break;
    }
    return new String(buf);
  }

  public int doCrc(byte[] buffer, int offset, int count)
  {
    /* Return the CRC of ptr[0]..ptr[count-1] */
    int crc = 0;
    for (int i = 0; i < count; i++)
    {
      crc = ((crc << 8) & 0xff00) ^ (CRCTABLE[(((crc & 0xff00) >> 8) ^ buffer[offset + i]) & 0xff] & 0xffff);
    }
    return crc;
  }

  void makeCrcTable()
  /* Fill the crctable[] array needed by doCrc */
  {
    int oneByte, bit;
    int crc;

    for (oneByte = 0; oneByte < 256; oneByte++)
    {
      crc = oneByte << 8;
      for (bit = 0; bit < 8; bit++)
      {
        crc = (((crc & 0x8000) != 0) ? (crc << 1) ^ 0x1021 : crc << 1);
      }
      CRCTABLE[oneByte] = crc;
    }
  }

  public void requestSend(String resourceName)
  {
    InputStream is = ADTPro.class.getClassLoader().getResourceAsStream(resourceName);
    if (is != null)
    {
      // Run this on a thread...
      _parent.setMainText(Messages.getString("CommsThread.4")); //$NON-NLS-1$
      _parent.setSecondaryText(resourceName); //$NON-NLS-1$
      _worker = new Worker(is);
      _worker.start();
    }
    else
    {
      System.out.println("Didn't find anything there.");
    }
  }

  public void requestStop()
  {
    _shouldRun = false;
    if (_worker != null) _worker.requestStop();
    try
    {
      _transport.close();
    }
    catch (Exception ex)
    {
      System.out.println("CommsThread.requestStop() exception: " + ex);
    }
  }

  public class Worker extends Thread
  {

    public Worker(InputStream is)
    {
      _is = is;
    }

    public void run()
    {
      // System.out.println("CommsThread.Worker.Run() starting.");
      try
      {
        char[] buffer = new char[_is.available()];
        InputStreamReader isr = new InputStreamReader(_is);
        isr.read(buffer);
        _parent.setProgressMaximum(buffer.length);
        _transport.setSlowSpeed(300);
        _startTime = new GregorianCalendar();
        for (int i = 0; i < buffer.length; i++)
        {
          if (_shouldRun == false) break;
          if (buffer[i] == 0x0d)
          {
            _transport.writeByte(0x8d);
            try
            {
              sleep(500);
            }
            catch (Exception e)
            {}
          }
          else
            if (buffer[i] != 0x0a) _transport.writeByte(buffer[i]);
          _parent.setProgressValue(i + 1);
        }
        _transport.pushBuffer();
      }
      catch (Exception e)
      {
        System.out.println(e);
      }
      _endTime = new GregorianCalendar();
      _diffMillis = (float) (_endTime.getTimeInMillis() - _startTime.getTimeInMillis()) / (float) 1000;
      _parent.setSecondaryText(Messages.getString("CommsThread.22") + " in " + _diffMillis + " seconds.");
      System.out.println("Text file sent in " + (float) (_endTime.getTimeInMillis() - _startTime.getTimeInMillis())
          / (float) 1000 + " seconds.");
      _transport.flushReceiveBuffer();
      _transport.setFullSpeed();
      // System.out.println("CommsThread.Worker.Run() stopping.");
    }

    public void requestStop()
    {
      _shouldRun = false;
    }

    InputStream _is;
  }

  public static int lastFileNumber = 0;
}
